/* Copyright 1996, Jan D. Wolter and Steven R. Weiss, All Rights Reserved. */

/* PARTIPATION FILE reader/writer.
 */

#include "backtalk.h"

#include <ctype.h>
#include <errno.h>
#include <time.h>
#include <sys/stat.h>

#include "sysdict.h"
#include "user.h"
#include "dict.h"
#include "str.h"
#include "sel.h"
#include "file.h"
#include "ps_part.h"
#include "lock.h"
#include "partdir.h"
#include "set_uid.h"

#ifdef HAVE_LIMITS_H
#include <limits.h>
#else
#define INT_MAX 32767
#endif

#define SECS_PER_DAY (60*60*24)

#ifdef DMALLOC
#include "dmalloc.h"
#endif


/* PARTICIPATION FILE - Normally one participation file is kept in memory.
 * It is for the conference in the sysdict variable "conf" and has been
 * read if "part_loaded" is true.
 *
 * "part_modified" is true if any change has been made to the data stored in
 * a PicoSpan-style participation file.  "ext_modified" is true if any of the
 * data added by Backtalk has been changed.  If either has been changed, we
 * should write the new one out before exiting.
 *
 * If "no_part" is true, we were unable to open a participation file for the
 * user (possibly because he is anonymous, possibly because of permission
 * problems) so we will run as if he was just joining for the first time, and
 * not attempt to save any updates.
 *
 * The participation file is stored in a linked list.
 *
 * Note that we hold the participation file open between loading it and
 * writing it, so that our lock on it doesn't go away.
 */

int part_loaded= FALSE;		/* Have we got participation file in memory? */
int part_type;			/* What type of particip. file was it?	 */
				/*   0= Picospan only (never happens)    */
				/*   1= Picospan plus extension file     */
				/*   2= Backtalk                         */

int part_modified= FALSE;	/* Need to write out part file when done? */
int ext_modified= FALSE;	/* Need to write out extensions when done? */

int snooping= FALSE;		/* Is conference open for snooping only? */

int no_part= FALSE;		/* Don't use a participation file */
time_t part_date= -1L;		/* Last mod date - -1=dunno, 0=brandnew */
FILE *part_fp= NULL;
FILE *ext_fp= NULL;
char part_name[BFSZ];		/* fullpath name of current particip. file */
char *alias= NULL;		/* User's full name, as in particip. file */
HashTable *conf_etc;		/* Miscellaneous conference fields */

struct pfe {
    int item_no;		/* Item number */
    int resp_no;		/* Number of responses read including item */
    time_t  date;		/* Time of last reading */
    char *my_title;		/* User's own title for the item */
    char forgotten;		/* Is this item forgotten? */
    char favorite;		/* Is this item a favorite? */
    char delete;		/* Should this entry not be written out? */
    HashTable *etc;		/* Other random fields */
    struct pfe *next;		/* Pointer to next entry */
    } *pflist;

struct pfe *last_part= NULL;	/* Last pflist item looked at */

void writepart(void);
void freepart(void);
int openpf(int snoop);
int loadpart(int snoop);
void newpart(int);
void writepspart(void);
void writebtpart(int ext);


/* CLOSEPART -- Close the currently open participation file if there is one,
 * writing it out if it has been modified.  If any item has been modified,
 * it is very important that this be called to write the pf out.
 */

void closepart()
{
    if (!part_loaded) return;

    if ((part_modified || ext_modified) && !snooping)
	writepart();
    else
	freepart();

    /* Locking is only on main file, so close any extension first */
    if (ext_fp != NULL)
    {
	fclose(ext_fp);
	ext_fp= NULL;
    }

    /* Close the real participation file */
    if (part_fp != NULL)
    {
	fclose(part_fp);
	part_fp= NULL;
	unlock(part_name);
    }

    pflist= NULL;
    last_part= NULL;

    part_loaded= FALSE;
    part_date= -1L;
}


/* REMOVEPART -- Close and delete the participation file of the currently open
 * conference.
 */

void removepart()
{
    int had_ext;

    if (no_part || !part_loaded) return;

    if (part_fp != NULL)
    {
	fclose(part_fp);
	part_fp= NULL;
	unlock(part_name);
    }

    if (had_ext= (ext_fp != NULL))
    {
	fclose(ext_fp);
	ext_fp= NULL;
    }

    if (!snooping)
    {
#if defined(PARTUTIL) && !defined(PART_DIR)
	partutil("-d",sdict(VAR_CONF),sdict(VAR_ID));
	if (had_ext)
	    partutil("-D",sdict(VAR_CONF),sdict(VAR_ID));
#else
	as_user();
	unlink(part_name);
	as_cfadm();

	if (had_ext)
	{
	    int pnlen= strlen(part_name);
	    part_name[pnlen]= 'x';
	    part_name[pnlen+1]= '\0';
	    unlink(part_name);
	    part_name[pnlen]= '\0';
	}
#endif
    }

    freepart();
    pflist= NULL;
    last_part= NULL;
    part_loaded= part_modified= ext_modified= FALSE;
    no_part= TRUE;
    part_date= -1L;
}


/* OPENPART - Make sure the participation file for the current conference has
 * been loaded in memory.  If the argument is true, we will open it in "snoop"
 * mode, in which participation files are read-only.  You don't normally have
 * to call this unless you want to snoop, because calling most any routine
 * that does anything to a participation file does call this.  Returns true if
 * a new participation file was created.
 */

int openpart(int snoop)
{
    if (part_loaded) return 0;
    snooping= snoop;

    if (sdict(VAR_ID)[0] == '\0' || sdict(VAR_HOMEDIR)[0] == '\0')
    {
	no_part= TRUE;
	return 0;
    }
    else
	return loadpart(snoop);
}


/* LOADPART -- Read in a user's participation file for the current conference
 * given directory name.  Create an empty one (in memory, so far) if he
 * doesn't have one.  If snoop is true, we don't create a participation
 * file.  Returns true if we created one.
 */

int loadpart(int snoop)
{
    char buf[BFSZ];
    int len;

#ifdef PART_DIR
    bbspartpath(part_name,sdict(VAR_ID),sdict(VAR_PARTICIP));
#else
    homepartpath(part_name,sdict(VAR_HOMEDIR),sdict(VAR_PARTICIP));
#endif

    /* Open, create, and lock the participation file */
    if (openpf(snoop))
    {
	newpart(!no_part);
	return 1;
    }
    else if (no_part)
	return 0;

    /* Read and check magic code number */
    if (fgets(buf,BFSZ,part_fp) == NULL)
    {
    	/* Ignore empty file */
	newpart(FALSE);
	return 1;
    }

    conf_etc= NULL;

    if (!strcmp(buf,"!<pr03>\n"))
    {
#if DEFAULT_PART_TYPE == 0
	/* Picospan only - we never do this */
	part_type= 0;
        return loadpspart();
#else
	/* Picospan with possible extension file */
	part_type= 1;
	if (loadpspart()) return 1;
	/* Try loading an extension file */
	if (!openpfext(snoop)) loadbtpart(TRUE);
	return 0;
#endif
    }

    if (!strcmp(buf,"!<pr10>\n"))
    {
	part_type= 2;
        return loadbtpart(FALSE);
    }

    if ((len= strlen(buf)) < 8 &&
        (!strncmp(buf,"!<pr03>",len) || !strncmp(buf,"!<pr10>",len)) )
    {
	/* Ignore truncated participation file */
	newpart(FALSE);
	return 1;
    }

    /* Bomb on things that can't have been part files */
    die("Participation file %s has bad magic number", part_name);
}


/* LOADPSPART -- Read in a Picospan-style participation file.  The file's name
 * is already in the global part_name, and it has already been opened on the
 * part_fp file pointer, and the first line has already been read in and
 * processed.  We just finish the job, returning 0 if we were able to read
 * the file, and 0 if we had to concoct a new one.
 */

int loadpspart()
{
    char buf[BFSZ], *t;
    struct pfe *npf,*lpf=NULL;

    /* Read alias name */
    if (fgets(buf,BFSZ,part_fp) == NULL)
    {
	/* EOF - Assume participation file was truncated */
	newpart(FALSE);
	return 1;
    }
    buf[BFSZ-2]= '\n'; buf[BFSZ-1]= '\0';
    if ((t= strchr(buf,'\n')) == NULL)
    {
	/* No terminating newline - Assume file was truncated */
	newpart(FALSE);
	return 1;
    }
    *t= '\0';
    alias= strdup(buf);
    sets(VAR_ALIAS,buf,SM_COPY);

    while (fgets(buf,BFSZ,part_fp) != NULL)
    {
	if (index(buf,'\n') == NULL)
	    /* Discard lines with no terminating newline */
	    continue;

	npf= (struct pfe *)malloc(sizeof(struct pfe));
#ifdef YAPP_COMPAT
	if (sscanf(buf,"%d %d %lx",&npf->item_no,
				   &npf->resp_no,
				   (unsigned long *)&npf->date) != 3)
	    die("syntax error in participation file %s: %s",part_name,buf);
	if (npf->forgotten= (npf->resp_no < 0))
	    npf->resp_no= -npf->resp_no;
#else
	/* Fancier scanning to handle case of "-0" responses */
	if (!isascii(buf[0]) || !isdigit(buf[0]))
	    die("syntax error in participation file %s: '%s'",part_name,buf);
	npf->item_no= atoi(buf);
	t= firstout(firstin(buf," ")," ");
	if (npf->forgotten= (*t == '-')) t++;
	if (sscanf(t,"%d %lx",&npf->resp_no,
			      (unsigned long *)&npf->date) != 2)
	    die("syntax error in participation file %s: %s",part_name,buf);
#endif /* YAPP_COMPAT */

	npf->delete= 0;
	npf->favorite= 0;
	npf->my_title= NULL;
	npf->etc= NULL;
	npf->next= NULL;
	if (lpf == NULL)
	    pflist= npf;
	else
	    lpf->next= npf;
	lpf= npf;
    }

    part_loaded= TRUE;
    part_modified= FALSE;
    ext_modified= FALSE;
    return 0;
}


/* NEW_ETC_INT - Store a new integer value in the given hashtable.  If the
 * value is zero, undefine it.
 */

void etc_int(HashTable **ht, char *tag, int value)
{
    int isnew;
    HashEntry *he;

    if (value == 0)
    {
	/* Undefine it */
	if (*ht == NULL) return;
	he= FindHashEntry(*ht, tag);
	if (he == NULL) return;
	DeleteHashEntry(he);
	return;
    }

    /* Create a hash table, if we didn't have one */
    if (*ht == NULL)
    {
	*ht= (HashTable *)malloc(sizeof(HashTable));
	InitHashTable(*ht);
    }
    /* Add or find the entry */
    he= CreateHashEntry(*ht, tag, &isnew);

    /* ...freeing any old value */
    if (!isnew) free_val(&(he->t));

    /* Store the new value */
    he->t.flag= TK_INTEGER;
    he->t.val= (void *)value;
}


/* NEW_ETC_STR - Store a new string value in the given hashtable.
 */

void etc_str(HashTable **ht, char *tag, char *value, int copy)
{
    int isnew;
    HashEntry *he;

    /* Create a hash table, if we didn't have one */
    if (*ht == NULL)
    {
	*ht= (HashTable *)malloc(sizeof(HashTable));
	InitHashTable(*ht);
    }
    /* Add or find the entry */
    he= CreateHashEntry(*ht, tag, &isnew);

    /* ...freeing any old value */
    if (!isnew) free_val(&(he->t));

    /* Store the new value */
    if (copy) value= strdup(value);
    he->t.flag= TK_STRING|TKF_FREE;
    he->t.val= (void *)value;
}


/* FIND_PF - Get a pointer to the structure describing the participation
 * file entry for the given item in the current conference.  Returns 0
 * if an entry existed before this call.  Returns 1 if no entry existed,
 * and all entries have lower item numbers.  Returns 2 if no entry existed,
 * but there are entries with higher item numbers.  If create is true, and
 * entry is created if none existed before.  Should never be called if
 * no_part is true.
 * 
 * This does some optimization to work faster when items are searched for
 * in sequential order, which they usually are.
 */

int find_pf(int item, int create, struct pfe **pf)
{
    struct pfe *pfp, *pfn;
    int rc;
	
    openpart(FALSE);

    /* Check if we found it on the last scan */
    if (last_part == NULL)
    {
	/* No previous scan - start scan from front of list */
	*pf= pflist;
    }
    else if (last_part->item_no == item)
    {
	/* Yup, no scan needed */
	*pf= last_part;
	return 0;
    }
    else if (last_part->item_no < item)
    {
	/* Nope, but item is after last one, so start scan from there */
	*pf= last_part;
    }
    else
    {
	/* Nope, old scan is useless - start from front of list */
	*pf= pflist;
    }

    /* walk the list, starting wherever */
    for (pfp= NULL;
	 *pf != NULL && (*pf)->item_no < item;
	 pfp= *pf, *pf= (*pf)->next)
    		;

    /* If we found it, get out of here */
    if (*pf != NULL && (*pf)->item_no == item)
    {
	last_part= *pf;
	return 0;
    }

    /* Item not in participation file... */
    rc= (*pf == NULL) ? 1 : 2;

    if (!create)
    {
	/* ...give it up */
	int rc= (*pf == NULL) ? 1 : 2;
	*pf= NULL;
	return rc;
    }

    /* ...put it in */
    pfn= (struct pfe *)malloc(sizeof(struct pfe));
    pfn->item_no= item;
    pfn->resp_no= 0;
    pfn->forgotten= 0;
    pfn->date= 0;
    pfn->favorite= 0;
    pfn->my_title= NULL;
    pfn->etc= NULL;
    pfn->delete= 0;
    pfn->next= *pf;

    if (pfp == NULL)
	pflist= pfn;
    else
	pfp->next= pfn;
    part_modified= TRUE;

    last_part= *pf= pfn;
    return rc;
}


/* LOADBTPART -- Read in a Backtalk-style participation file.  The file's name
 * is already in the global part_name, and it has already been opened on the
 * part_fp file pointer, and the first line has already been read in and
 * processed.  We just finish the job, returning 0 if we were able to read
 * the file, and 0 if we had to concoct a new one.  If ext is true, load an
 * extension file instead, assuming we have already loaded the corresponding
 * Picospan-style participation file.
 */

int loadbtpart(int ext)
{
    char buf[BFSZ], *var, *val, *ch, *save;
    int save_size, save_i;
    long ival;
    struct pfe *npf= NULL, *lpf= NULL;
    int is_string, skip= FALSE;
    FILE *fp= ext ? ext_fp : part_fp;

    /* Set up our growable buffer for reading in strings */
    save= (char *)malloc(BFSZ);
    save_size= BFSZ;

    while (fgets(buf,BFSZ,fp) != NULL)
    {
	/* Lines starting with numbers point to a new item number. */
    	if (isascii(buf[0]) && isdigit(buf[0]))
	{
	    if (ext)
	    {
		/* Loading extension - find existing pf entry for the item.
		 * If there is none, skip this entire item, don't create one.
		 */
		skip= find_pf(atoi(buf), FALSE, &npf);
	    }
	    else
	    {
		/* Loading standard file - create a the entry */
		npf= (struct pfe *)malloc(sizeof(struct pfe));

		npf->item_no= atoi(buf);
		npf->delete= 0;
		npf->favorite= 0;
		npf->forgotten= 0;
		npf->my_title= NULL;
		npf->resp_no= 0;
		npf->date= 0;
		npf->etc= NULL;
		npf->next= NULL;

		if (pflist == NULL)
		    pflist= npf;
		else
		    lpf->next= npf;

		lpf= npf;
		skip= FALSE;
	    }
	    continue;
	}
	else if (skip)
	    continue;

	/* Ignore blanks */
	if (buf[0] == '\n') continue;

	/* Find start of value */
	if ((val= strchr(buf,'=')) != NULL)
	{
	    *val= '\0';
	    *val++;
	    is_string= (val[0] == '(');
	}
	else
	{
	    char *t= strchr(buf,'\n');
	    if (t != NULL) *t= '\0';
	    is_string= 0;
	}

	/* Save variable name */
	var= strdup(buf);

	if (is_string)
	{
	    /* Read string values - can run across multiple lines */
	    int esc= 0;
	    save_i= 0;
	    ch= val+1;
	    for (;;)
	    {
	    	if (ch == '\0')
		{
		    /* Long line broken by fgets or truncated file */

		    if (fgets(buf,BFSZ,fp) == NULL)
			break;  /* truncated - stop */
		    
		    ch= buf;
		}

		if (*ch == ')' && !esc)
		    break;	/* Normal end of string */
		
		if (esc= (*ch == '\\' && !esc))
		{
		    ch++;
		    continue;
		}

		if (save_i+1 >= save_size)
		{
		    save_size+= BFSZ;
		    save= (char *)realloc(save, save_size);
		}
		save[save_i++]= *(ch++);
	    }
	    save[save_i]= '\0';
	}
	else if (val == NULL)
	    /* Boolean value */
	    ival= 1;
	else if (isascii(val[0]) && (isdigit(val[0]) || val[0] == '-'))
	    /* Integer value */
	    ival= atoi(val);
	else
	    die("syntax error in participation file %s: %s",part_name,buf);

	/* Handle special fields */
	if (npf == NULL)
	{
	    if (!strcmp(var,"alias"))
	    {
		if (is_string)
		{
		    alias= strdup(save);
		    sets(VAR_ALIAS,save,SM_COPY);
		}
	    }
	    else if (is_string)
		etc_str(&conf_etc,var,save,TRUE);
	    else
		etc_int(&conf_etc,var,ival);
	}
	else
	{
	    if (!strcmp(var,"r"))
	    {
	    	/* This gets much fancier in the future */
		if (!is_string) npf->resp_no= ival;
	    }
	    else if (!strcmp(var,"d"))
	    {
		if (!is_string) npf->date= (time_t)ival;
	    }
	    else if (!strcmp(var,"forgot"))
	    {
		if (!is_string) npf->forgotten= ival;
	    }
	    else if (!strcmp(var,"fav"))
	    {
		if (!is_string) npf->favorite= ival;
	    }
	    else if (!strcmp(var,"title"))
	    {
		if (is_string) npf->my_title= strdup(save);
	    }
	    else if (is_string)
		etc_str(&(npf->etc),var,save,TRUE);
	    else
		etc_int(&(npf->etc),var,ival);
	}
	free(var);
    }

    part_loaded= TRUE;
    part_modified= FALSE;
    ext_modified= FALSE;
    free(save);
    return 0;
}


/* NEWPART -- Create a new participation file for the conference in memory.
 * If the "newuser" flag is true, we construct a participation file with some
 * unseen items and some new items, as determined by the "new_isel" and
 * "new_since" settings in the conference settings file.  Otherwise, we
 * just make all items new.
 */

void newpart(int newuser)
{
    pflist= NULL;
    last_part= NULL;

    /* Set alias name */
    alias= strdup(user_fname());
    sets(VAR_ALIAS,alias,SM_COPY);

    part_loaded= TRUE;
    part_modified= !no_part;
    ext_modified= FALSE;
    part_type= DEFAULT_PART_TYPE;
    conf_etc= NULL;

    if (newuser)
    {
    	char new_isel[BFSZ+1];
    	int new_since, item, nresp, flags;
    	time_t idate, rdate, sdate, now;
	struct pfe *pfn, *pfp;

	part_date= 0L;

	/* Get instructions for which should be new from the settings file  */
	get_settings(sdict(VAR_CONFDIR),new_isel,&new_since);

	/* Everything is new - do nothing */
	if (!strcmp(new_isel,"1-$")) return;

	now= time(0L);
	sdate= now - new_since * SECS_PER_DAY;

	pfn= pfp= NULL;
	for (seek_item(1); !next_item(&item,&nresp,&flags,&idate,&rdate); )
	{
	    pfn= (struct pfe *)malloc(sizeof(struct pfe));
	    pfn->item_no= item;
	    pfn->delete= 0;
	    pfn->forgotten= 0;
	    pfn->favorite= 0;
	    pfn->resp_no= 0;
	    pfn->my_title= NULL;
	    pfn->etc= NULL;

	    if ((new_since > 0 && rdate >= sdate) || in_sel(item,new_isel))
		pfn->date= 0;	/* Item New */
	    else
		pfn->date= now;	/* Item Unseen */

	    pfn->next= NULL;
	    if (pfp == NULL)
	    	pflist= pfn;
	    else
		pfp->next= pfn;
	    pfp= pfn;
	}

	/* Special handling for last item and $ sign in isels */
	if (pfn != NULL && pfn->date != 0 && strchr(new_isel,'$') != NULL)
	    pfn->date= 0;
    }
}


/* LAST_READ -- Return the number of the last response to item that the user
 * has read.  Return resp_no = -1 if it is unread.  The flag forgotten is
 * set to true if the item has been forgotten.
 */

void last_read(int item, int *resp_no, int *forgotten, time_t *date)
{
    struct pfe *pf;
	
    /* Set default values */
    *forgotten= FALSE;
    *resp_no= -1;
    *date= 0;

    if (no_part) return;

    if (find_pf(item, FALSE, &pf))
	/* No such item */
	return;

    /* found item */
    *forgotten= pf->forgotten;
    *resp_no= pf->resp_no-1;
    *date= pf->date;
}


/* UPDATE_ITEM -- Mark an item read up through the given response.  If the
 * response is -1, the entire item is unseen.  If the response is -2, forget
 * the item.  If response is -3, don't update the response number -- only
 * update the time.  If the response is -4, remember (unforget) the item.
 * If the item is forgotten, marking it seen does not unforget it.
 * This version is fancied up so that multiple calls on the same item in
 * rapid succession can be handled efficiently.  Unless rewind is set, the
 * this will only increase the number of responses seen.  If resp is -1, it
 * will not be changed.
 */

void update_item(int item, int resp, time_t now, int rewind)
{
    struct pfe *pf;

    if (no_part) return;

    if (find_pf(item, TRUE, &pf))
    {
	/* A newly created participation file entry */
	if (resp == -2)
	{
	    /* forget an unseen item */
#ifdef YAPP_COMPAT
	    pf->resp_no= 1;
#else
	    pf->resp_no= 0;
#endif
	    pf->forgotten= 1;
	    pf->date= 0;
	}
	else
	{
	    /* See the item */
	    pf->resp_no= (resp < 0) ? 0 : resp+1;
	    pf->forgotten= 0;
	    pf->date= now;
	}
	part_modified= TRUE;
    }
    else
    {
	/* Update existing entry */
	if (resp == -2 || resp == -4)
	{
	    /* Remembering and Forgetting */
	    if ((resp == -4) == (pf->forgotten)) 
	    {
#ifdef YAPP_COMPAT
		/* Yapp participation file format can't handle forgetting
		 * unread items so mark the first response read in that case.
		 */
		if (pf->resp_no == 0) pf->resp_no= 1;
#endif
		pf->forgotten= !pf->forgotten;
		part_modified= TRUE;
		/* If we are forgetting, automatically disfavor the item */
		if (resp == -2 || pf->favorite)
		{
		    pf->favorite= 0;
		    ext_modified= TRUE;
		}
	    }
	}
	else
	{
	    /* Plain old looking at things */
	    if (resp != -3 && (rewind || pf->resp_no <= resp))
	    {
		pf->resp_no= resp+1;
		/* Disfavor temporary favorites */
		if (pf->favorite == 2)
		{
		    pf->favorite= 0;
		    ext_modified= TRUE;
		}
	    }
	    pf->date= now;
	    part_modified= TRUE;
	}
    }
}


/* PF_CONF_SETI - Set one of the the conference-wide, post-Picospan fields in
 * the participation file to an integer value.  This is the main entry point
 * for other modules to call if setting a value.
 */

void pf_conf_seti(char *tag, int value)
{
    if (no_part) return;
    etc_int(&(conf_etc), tag, value);
    ext_modified= TRUE;
}


/* PF_CONF_SETS - Set one of the the conference-wide, post-Picospan fields in
 * the participation file to an string value.  This is the main entry point
 * for other modules to call if setting a value.
 */

void pf_conf_sets(char *tag, char *value, int copy)
{
    if (no_part)
    {
	if (!copy) free(value);
	return;
    }

    etc_str(&(conf_etc), tag, value, copy);
    ext_modified= TRUE;
}


/* PF_ITEM_SETI - Set one of the post-Picospan fields in the participation file
 * entry for an item to an integer value.
 */

void pf_item_seti(int item, char *tag, int value)
{
    struct pfe *pf;

    if (no_part) return;

    find_pf(item, TRUE, &pf);

    if (!strcmp(tag,"fav"))
	pf->favorite= value;
    else
	etc_int(&(pf->etc), tag, value);

    ext_modified= TRUE;
}


/* PF_ITEM_SETS - Set one of the post-Picospan fields in the participation file
 * entry for an item to a string value.  If 'copy' is false, the string is
 * not already in malloced memory.
 */

void pf_item_sets(int item, char *tag, char *value, int copy)
{
    struct pfe *pf;

    if (no_part)
    {
	if (!copy) free(value);
	return;
    }

    find_pf(item, TRUE, &pf);

    if (!strcmp(tag,"title"))
	pf->my_title= value;
    else
	etc_str(&(pf->etc), tag, value, copy);

    ext_modified= TRUE;
}

/* PF_CONF_GET - Get a value from the header.  Returns true if not defined.
 */

int pf_conf_get(char *tag, Token *val)
{
    if (!strcmp(tag,"alias"))
    {
    	val->val= (void *)strdup(alias);
	val->flag= TK_STRING | TKF_FREE;
	return 0;
    }
    else if (conf_etc == NULL)
	return 1;
    else
    {
	HashEntry *h= FindHashEntry(conf_etc, tag);
	if (h == NULL)
	    return 1;
	*val= h->t;
	uniquify_token(val);
	return 0;
    }
}


/* PF_ITEM_GET - Get a value from an item in the participation file.  Returns
 * true if not defined.
 */

int pf_item_get(int item, char *tag, Token *val)
{
    struct pfe *pf;
    find_pf(item, FALSE, &pf);
    if (pf == NULL) return 1;

    if (!strcmp(tag,"title"))
    {
    	val->val= (void *)strdup(pf->my_title);
	val->flag= TK_STRING | TKF_FREE;
	return 0;
    }
    else if (!strcmp(tag,"fav"))
    {
    	val->val= (void *)(int)pf->favorite;
	val->flag= TK_INTEGER;
	return 0;
    }
    else if (!strcmp(tag,"r"))
    {
    	val->val= (void *)pf->resp_no;
	val->flag= TK_INTEGER;
	return 0;
    }
    else if (!strcmp(tag,"d"))
    {
    	val->val= (void *)pf->date;
	val->flag= TK_TIME;
	return 0;
    }
    else if (!strcmp(tag,"forgot"))
    {
    	val->val= (void *)(int)pf->forgotten;
	val->flag= TK_INTEGER;
	return 0;
    }
    else if (pf->etc == NULL)
	return 1;
    else
    {
	HashEntry *h= FindHashEntry(pf->etc, tag);
	if (h == NULL)
	    return 1;
	*val= h->t;
	uniquify_token(val);
	return 0;
    }
}

/* PF_RFAV -- update an item's "fav" note based on the current setting of the
 * rfav flag.
 */

void pf_rfav(int item)
{
    struct pfe *pf;

    if (no_part) return;

    switch (idict(VAR_RFAV))
    {
    case 0:
	break;

    case 1:
	find_pf(item, TRUE, &pf);
	pf->favorite= 1;
	ext_modified= TRUE;
	break;
	
    case 2:
	find_pf(item, TRUE, &pf);
	if (pf->favorite != 1)
	{
	    ext_modified= TRUE;
	    pf->favorite= 2;
	}
	break;
    }
}


/* LOST_ITEMS -- The items between first and last are missing.  Delete them from
 * the partipation file.  Really just mark them for deletion.  If last is -1,
 * then delete through the end of the conference.
 */

void lost_items(int first,int last)
{
    struct pfe *pf;
    int state=1;

    if (no_part || last < first) return;

    openpart(FALSE);

    if (no_part) return;

    if (last == -1) last= INT_MAX;
	
    for (pf= pflist; pf != NULL; pf= pf->next)
    {
	if (state == 1 && pf->item_no >= first)
	    state= 2;
		
	if (state == 2)
	{
	    if (pf->item_no > last) break;

#if 0
	    /* If an item was deleted by Picospan, its index file
	     * may still be there.  But the following test means
	     * every user whoever saw the item before it was
	     * deleted would try to delete the index file once.
	     * That's a lot of overhead.  A cron script cleaning
	     * up leftover index files is a better solution.
	     */
	    if (!pf->delete)
		remove_index(sdict(VAR_CONFDIR),pf->item_no);
#endif
	    pf->delete= 1;
	}
    }
}


/* WRITEPART -- Write out the participation file.  A loadpart() or newpart()
 * call must preceed this.  This deallocates the memory that it was stored
 * in as a side effect.
 */

void writepart()
{
    struct pfe *pf, *npf;

    if (no_part)
    {
	if (part_loaded) freepart();
	return;
    }

#ifdef CONVERT_PART
#if DEFAULT_PART_TYPE != 1
    if (part_type == 1 && ext_fp != NULL)
    {
	/* If converting from a format with extension file, remove it */
	int pnlen= strlen(part_name);
	part_name[pnlen]= 'x';
	part_name[pnlen+1]= '\0';
	unlink(part_name);
	part_name[pnlen]= '\0';
	fclose(ext_fp);
	ext_fp= NULL;
    }
#endif
    part_type= DEFAULT_PART_TYPE;
#endif

    if (part_type == 0)
	writepspart();
    else if (part_type == 1)
    {
	if (ext_modified)
	{
	    if (ext_fp == NULL)
		openpfext(FALSE);
	    else
		rewind(ext_fp);
	    if (ext_fp != NULL)
		writebtpart(TRUE);
	}
	if (part_modified)
	    writepspart();
	else
	    freepart();
    }
    else
	writebtpart(FALSE);
}


/* WRITEPSPART -- Write out a PicoSpan-style participation file.  This
 * deallocates the memory that it was stored in as a side effect.
 */

void writepspart()
{
    struct pfe *pf, *npf;
    rewind(part_fp);

    fputs("!<pr03>\n",part_fp);
    fputs(alias,part_fp);
    fputc('\n',part_fp);
    free(alias);

    if (conf_etc != NULL)
    {
	DeleteHashTable(conf_etc);
	free(conf_etc);
	conf_etc= NULL;
    }

    for (pf= pflist; pf != NULL; pf= npf)
    {
	if (!pf->delete)
	{
#ifdef YAPP_COMPAT
	    fprintf(part_fp,"%d %d %lX\n",
		    pf->item_no,
		    pf->resp_no * (pf->forgotten ? -1 : 1),
		    (unsigned long)pf->date);
#else
	    /* This version allows a "-0" to be printed, which Picospan
	     * understands just fine, but which Yapp considers freakish.
	     */
	    fprintf(part_fp, pf->forgotten ? "%d -%d %lX\n" : "%d %d %lX\n",
		    pf->item_no,
		    pf->resp_no,
		    (unsigned long)pf->date);
#endif /* YAPP_COMPAT */
	    if (pf->my_title != NULL)
		free(pf->my_title);
	    if (pf->etc != NULL)
	    {
		DeleteHashTable(pf->etc);
		free(pf->etc);
	    }
	}
	npf= pf->next;
	free(pf);
    }

    /* We need this here because the file can get smaller, and we can't
     * destroy and recreate it because we'd lose our lock.
     */

    ftruncate(fileno(part_fp),(size_t)ftell(part_fp));
}


/* WRITEBTPART -- Write out a Backtalk-style participation file.  This
 * deallocates the memory that it was stored in as a side effect.
 *
 * If ext is true, then write an extenstion file instead, and do not deallocate
 * anything.
 */

void writebtpart(int ext)
{
    HashEntry *h;
    HashSearch hs;
    struct pfe *pf,*npf;
    FILE *fp= ext ? ext_fp : part_fp;

    rewind(fp);

    if (ext)
	fputs("!<px01>\n",fp);
    else
    {
	fputs("!<pr10>\n",fp);
	wrtstrtag(fp,"alias",alias);
	free(alias);
    }

    if (conf_etc != NULL)
    {
	for (h= FirstHashEntry(conf_etc, &hs); h != NULL; h= NextHashEntry(&hs))
	    wrttag(fp, h->key, &(h->t));
	if (!ext)
	{
	    DeleteHashTable(conf_etc);
	    free(conf_etc);
	    conf_etc= NULL;
	}
    }

    for (pf= pflist; pf != NULL; pf= npf)
    {
	if (!pf->delete)
	{
	    if (ext)
	    {
		if (pf->favorite || pf->my_title != NULL || pf->etc != NULL)
		    fprintf(fp,"%d\n",pf->item_no);
	    }
	    else
	    {
		fprintf(fp,"%d\nr=%d\nd=%ld\n",
			pf->item_no,
			pf->resp_no,
			(unsigned long)pf->date);
		if (pf->forgotten) fputs("forgot\n",fp);
	    }
	    if (pf->favorite == 1)
		fputs("fav\n",fp);
	    else if (pf->favorite)
		fprintf(fp, "fav=%d\n",pf->favorite);
	    if (pf->my_title != NULL)
	    {
		wrtstrtag(fp, "title", pf->my_title);
		if (!ext) free(pf->my_title);
	    }
	    if (pf->etc != NULL)
	    {
		for (h= FirstHashEntry(pf->etc, &hs); h != NULL;
		     h= NextHashEntry(&hs))
		    wrttag(fp, h->key, &(h->t));
		if (!ext)
		{
		    DeleteHashTable(pf->etc);
		    free(pf->etc);
		}
	    }
	}
	npf= pf->next;
	if (!ext) free(pf);
    }

    /* We need this here because the file can get smaller, and we can't
     * destroy and recreate it because we'd lose our lock.
     */

    ftruncate(fileno(fp),(size_t)ftell(fp));
}


/* FREEPART -- Deallocate the memory we have the participation file stored
 * in.
 */

void freepart()
{
    struct pfe *pf, *npf;

    free(alias);
    if (conf_etc != NULL)
    {
	DeleteHashTable(conf_etc);
	free(conf_etc);
	conf_etc= NULL;
    }

    for (pf= pflist; pf != NULL; pf= npf)
    {
	if (pf->my_title != NULL)
	    free(pf->my_title);
	if (pf->etc != NULL)
	{
	    DeleteHashTable(pf->etc);
	    free(pf->etc);
	}
	npf= pf->next;
	free(pf);
    }
}


/* HOMEPARTPATH -- Construct the full path name of the participation file,
 * given the file name.  Save the result in the passed-in buffer.
 * Returns a pointer to "path" -- Normal Version, but also used with Yapp 3.0
 * to find .cflist files.
 */

char *homepartpath(char *path, char *dir, char *name)
{
    int homelen;

    strcpy(path,dir);
    homelen= strlen(path);
#ifdef ALLOW_CFDIR
    strcpy(path+homelen,"/.cfdir/");
    if (access(path,0))
	strcpy(path+homelen+1,name);
    else
	strcat(path+homelen,name);
#else
    path[homelen]= '/';
    strcpy(path+homelen+1,name);
#endif /* ALLOW_CFDIR */
    return(path);
}


/* OPENPF - Open the current participation file and store the filepointer in
 * the part_fp global variable.  If there was no previous participation file
 * return 1, otherwise return 0.  If snoop is true, never create the file.
 */

int openpf(int snoop)
{
    no_part= FALSE;

    /* If we are in direct_exec mode, all this must be done as the user */
    as_user();

    /* Try opening an existing particiption file */
    part_fp= fopen(part_name, snoop ? "r" : "r+");

#ifdef PART_DIR
    if (!snoop && part_fp == NULL && errno == ENOENT)
    {
	/* User directory missing - create it */
	mkpartdir(part_name);
	part_fp= fopen(part_name, "r+");
    }
#else

#ifdef PARTUTIL
    /* If we fail because of permissions, try using partutil to fix them */
    if (part_fp == NULL && errno == EACCES)
    {
	partutil("-p",sdict(VAR_CONF),sdict(VAR_ID));
	part_fp= fopen(part_name, snoop ? "r" : "r+");
    }
#endif /* PARTUTIL */
#endif /* !PART_DIR */

    /* If we opened file, lock it and return */
    if (part_fp != NULL)
    {
	if (!no_part)
	    lock_exclusive(fileno(part_fp),part_name);
	as_cfadm();
	return 0;
    }
    else if (snoop)
    {
	as_cfadm();
	return 1;
    }

    /* If file exists but isn't readable then don't try to create it and
     * just run without one.
     */
    if (errno == EACCES)
    {
	no_part= TRUE;
	as_cfadm();
	return 1;
    }

    /* File doesn't exist - create one */
#ifdef PARTUTIL
    partutil("-c",sdict(VAR_CONF),sdict(VAR_ID));
#endif
    no_part= (no_part || ((part_fp= fopen(part_name,"w+")) == NULL));
    if (!no_part)
        lock_exclusive(fileno(part_fp),part_name);
    as_cfadm();
    return 1;
}


/* OPENPFEXT - Open the extension file to the current participation file and
 * store the filepointer in the ext_fp global variable.  If there was no
 * previous participation file return 1, otherwise return 0.  If snoop is
 * true, never create the file.  The main participation file should already
 * be open.
 */

int openpfext(int snoop)
{
    int pnlen= strlen(part_name);

    /* Append an x to partname */
    part_name[pnlen]= 'x';
    part_name[pnlen+1]= '\0';

    /* If we are in direct_exec mode, all this must be done as the user */
    as_user();

    /* Try opening an existing extension file */
    ext_fp= fopen(part_name, snoop ? "r" : "r+");

#ifdef PARTUTIL
    /* If we fail because of permissions, try using partutil to fix them */
    if (ext_fp == NULL && errno == EACCES)
    {
	partutil("-P",sdict(VAR_CONF),sdict(VAR_ID));
	ext_fp= fopen(part_name, snoop ? "r" : "r+");
    }
#endif /* PARTUTIL */

    /* If we opened file, return (don't lock) */
    if (ext_fp != NULL || snoop)
    {
	part_name[pnlen]= '\0';
	if (ext_fp != NULL)
	{
	    char buf[BFSZ];
	    /* If the file is empty, go ahead and use it, but if it has
	     * something other than the normal header, decline to use it.
	     * */
	    if (fgets(buf,BFSZ,ext_fp) != NULL &&
		 strcmp(buf,"!<px01>\n"))
	    {
		/* weird file - close it */
		fclose(ext_fp);
		ext_fp= NULL;
	    }
	}
        part_name[pnlen]= '\0';
	as_cfadm();
	return (ext_fp == NULL);
    }

    /* If file exists but isn't readable then don't try to create it and
     * just run without one.
     */
    if (errno == EACCES)
    {
	part_name[pnlen]= '\0';
	as_cfadm();
	return 1;
    }

    /* File doesn't exist - create one */
#ifdef PARTUTIL
    partutil("-C",sdict(VAR_CONF),sdict(VAR_ID));
#endif
    ext_fp= fopen(part_name,"w+");

    part_name[pnlen]= '\0';
    as_cfadm();
    return 1;
}


#ifdef PARTUTIL

#include "waittype.h"

/* PARTUTIL - execute the partutil program to create, permit, or destory
 * user's participation file for a conference.  Flag should be "-c" to
 * create, "-p" to permit, or "-d" to destroy a participation file.  A flag
 * of "-l" permits (or creates, if necessary) a .cflist for the user.  In
 * this case conf should be NULL.
 */

int partutil(char *flag, char *conf, char *user)
{
    pid_t cpid,wpid;
    waittype status;

    /* If we are executing directly, we don't need partutil */
    if (direct_exec) return 0;

    if ((cpid= fork()) == 0)
    {
	/* Child process: do the exec */
	if (conf == NULL)
	    execl(PARTUTIL,"partutil",flag,user,(char *)NULL);
	else
	    execl(PARTUTIL,"partutil",flag,conf,user,(char *)NULL);
	exit(127);
    }

    while ((wpid= wait(&status)) != -1 && wpid != cpid)
	;

    return (WIFEXITED(status)) ? WEXITSTATUS(status) : 127;
}
#endif /* PARTUTIL */


/* SET_CONF_ALIAS - Change the default full name for the user for the
 * currently open conference.  "newname" should be malloc'ed memory.
 * Aliases with newlines are clipped at the newline.
 */

void set_conf_alias(char *newname)
{
    char *p;
    if ((p= strchr(newname,'\n')) != NULL) *p= '\0';
    if (alias != NULL) free(alias);
    alias= newname;
    part_modified= TRUE;
}

/* PARTDATE - Return the last modification date of the currently conference's
 * participation file.  If it didn't exist before this session, return 0.
 */

time_t partdate()
{
    struct stat st;
    if (!part_loaded) return 0L;
    if (part_date == -1L)
	part_date= stat(part_name,&st) ? 0L : st.st_mtime;
    return part_date;
}


/* FAV_SEL -- Given an item selector, split it into selectors for favorite,
 * brandnew, and newresp items based on participation file info.  Returns
 * up to three pointers to strings in malloced memory.  Pointers may be NULL.
 * if there is nothing in a category.
 *    0 - no-op - just return what was passed in.
 *    1 - return favorites and others
 *    2 - return brandnew and newresp
 *    3 - return favorites, non-favorite brandnew, and non-favorite newresp.
 * If you call it with flag=0, you just get the copy of the original selector,
 * but you don't want to do that, because it is slow.  Function returns a
 * count of things returned.  Passed in string is mangled.
 */

#define FF_LASTITEM 2147483647

long ff_inum(char *s)
{
    return (*s == '$') ? FF_LASTITEM : atoi(s);
}

struct ff_tmp {
    char *sel;		/* A selector constructed so far */
    int sz;		/* Amount of memory allocated for sel */
    int i;		/* Our current index in sel */
    long first;		/* Begining of current range */
    long last;		/* End of current range - negative if none */
};


/* Append an item to the selector.  Call with item=-1 to wrap up.  */

void ff_app(struct ff_tmp *s, long item)
{
    /* Never been called before - initialize */
    if (s->sel == NULL)
    {
	if (item == -1) return;
	s->sel= (char *)malloc(s->sz= BFSZ);
	s->i= 0;
	s->first= s->last= item;
	return;
    }

    /* Check if we are extending the current range */
    if (s->last > 0 && (item == s->last + 1 || item == FF_LASTITEM))
    {
	s->last= item;
	return;
    }

    if (s->last == -1)
    {
	s->first= s->last= item;
	return;
    }

    /* Make sure we have some memory */
    if (s->sz - s->i <= 25)
	s->sel= (char *)realloc(s->sel, s->sz+= BFSZ);

    /* Write out a completed range */
    if (s->last > 0)
    {
	if (s->i > 0) s->sel[(s->i)++]= ',';
	if (s->first == s->last)
	    sprintf(s->sel + s->i, "%d", s->first);
	else if (s->last == FF_LASTITEM)
	    sprintf(s->sel + s->i, "%d-$", s->first);
	else
	    sprintf(s->sel + s->i, "%d-%d", s->first, s->last);
	s->i+= strlen(s->sel + s->i);
    }

    /* Start a new range */
    s->first= s->last= item;
}


int fav_sel(char *isel, int flag, char **is1, char **is2, char **is3)
{
    char *next_range, *dash;
    int reverse, rc, favi,newi,otheri, len;
    long i,i1,i2;
    struct pfe *pf;
    struct ff_tmp fav, new, other;
    struct ff_tmp *pnew= (flag & 2) ? &new : &other;

    /* Flag zero: just return what we were give */
    if (flag == 0) {*is1 = isel; return 1;}

    /* Empty isel: return a lot of empty stuff */
    if (isel == NULL || *isel == '\0')
    {
	*is1= *is2= *is3= NULL;
	return (flag==3 ? 2 : 3);
    }

    fav.sel= new.sel= other.sel= NULL;

    for (;;)
    {
	/* First parse into comma-separated hunks */
	next_range= strchr(isel,',');
	if (next_range != NULL)
	    *(next_range++)= '\0';

	/* Get begining and end of range */
	dash= strchr(isel,'-');
	if (dash == NULL)
	{
	    /* Isolated $ will be converted into current last item number */
	    i1= i2= (*isel == '$') ? idict(VAR_MAXITEM) : atoi(isel);
	}
	else
	{
	    *dash= '\0';
	    i1= ff_inum(isel);
	    i2= ff_inum(dash+1);
	}

	/* Reverse range if it is non-increasing */
	if (i1 > i2) 
	{
	    reverse= i1;
	    i1= i2;
	    i2= reverse;
	    favi= (fav.sel == NULL) ? 0 : fav.i;
	    newi= (new.sel == NULL) ? 0 : new.i;
	    otheri= (other.sel == NULL) ? 0 : other.i;
	}
	else
	    reverse= 0;

	/* Walk the range */
	for (i= i1; i <= i2; i++)
	{
	    rc= find_pf(i, FALSE, &pf);
	    if (rc == 1)
	    {
		/* end of participation file - anything after it is brandnew */
		if (i < i2 || pnew->sel == NULL || pnew->first == -1)
		    ff_app(pnew,i);
		pnew->last= i2;
		break;
	    }
	    else if ((flag & 1) && rc == 0 && pf->favorite)
	    {
		/* found a favorite */
		ff_app(&fav,i);
	    }
	    else if ((flag & 2) && (rc == 2 || pf->resp_no == 0))
	    {
		/* found a brandnew */
		ff_app(&new,i);
	    }
	    else
	    {
		/* found something else */
		ff_app(&other,i);
	    }
	}

	/* Finish the fav, new, and other selectors up */
	if (flag & 1) ff_app(&fav,-1);
	if (flag & 2) ff_app(&new,-1);
	ff_app(&other,-1);

	/* Reverse the range, if needed */
	if (reverse)
	{
	    if ((flag & 1) && fav.sel != NULL)
	    {
		if (fav.sel[favi] == ',') favi++;
		rev_sel(fav.sel+favi);
	    }
	    if ((flag & 2) && new.sel != NULL)
	    {
		if (new.sel[newi] == ',') newi++;
		rev_sel(new.sel+newi);
	    }
	    if (other.sel != NULL)
	    {
		if (other.sel[otheri] == ',') otheri++;
		rev_sel(other.sel+otheri);
	    }
	}

	/* Get ready for the next iteration */
	if (next_range == NULL) break;
	isel= next_range;
    }

    /* Return things in suitable slots */
    switch (flag)
    {
    case 1:
	*is1= fav.sel;
	*is2= other.sel;
	*is3= NULL;
	return 2;

    case 2:
	*is1= new.sel;
	*is2= other.sel;
	*is3= NULL;
	return 2;

    case 3:
	*is1= fav.sel;
	*is2= new.sel;
	*is3= other.sel;
	return 3;
    }
}
